from .rcmod import *
from .utils import *
from .palettes import *
from .linearmodels import *
from .distributions import *
from .timeseries import *
from .matrix import *
from .miscplot import *
from .axisgrid import *
from .xkcd_rgb import xkcd_rgb
set()

__version__ = "0.5.0"
